#include "kint.hpp"
#include <area.hpp>
#include <entry.hpp>
#include <srarea.hpp>

extern "C" processor_t LPH;

static void mark_header( void ) {
   doASCI(0, 6);
   set_name(0,"K_MAGIC",0);
   set_name(6,"K_MajVer",0);
   set_name(7,"K_MinVer",0);
   op_hex(8, 0);
   set_name(8,"K_CSIZE",0);
   op_hex(0xc, 0);
   set_name(0xc,"K_DSIZE",0);
   op_hex(0x10, 0);
   set_offset(0x10,0,0);
   set_name(0x10,"STACK_BYTES",0);
   set_offset(0x14,0,0);
   set_name(0x14,"FloatPStack0",0);
   set_offset(0x18,0,0);
   set_name(0x18,"FloatPStack1",0);
   set_offset(0x1c,0,0);
   set_name(0x1c,"FloatPStack2",0);
   set_offset(0x20,0,0);
   set_name(0x20,"FloatPStack3",0);
   set_offset(0x24,0,0);
   set_name(0x24,"FloatPStack4",0);
   set_offset(0x28,0,0);
   set_name(0x28,"FloatPStack5",0);
   set_offset(0x2c,0,0);
   set_name(0x2c,"FloatPStack6",0);
   set_offset(0x30,0,0);
   set_name(0x30,"FloatPStack7",0);
   set_offset(0x34,0,0);
   set_name(0x34,"OnNullFunction1",0);
   set_offset(0x38,0,0);
   set_name(0x38,"OnNullFunction2",0);
   set_offset(0x3c,0,0);
   set_name(0x3c,"OnScreenChange",0);
   set_offset(0x40,0,0);
   set_name(0x40,"OnAlphaKeyDown",0);
   set_offset(0x44,0,0);
   set_name(0x44,"OnAlphaKeyUp",0);
   set_offset(0x48,0,0);
   set_name(0x48,"OnBeginStartup",0);
   set_offset(0x4c,0,0);
   set_name(0x4c,"OnEndStartup",0);
   set_offset(0x50,0,0);
   set_name(0x50,"OnWindowHidden",0);
   set_offset(0x54,0,0);
   set_name(0x54,"OnWindowMove",0);
   set_offset(0x58,0,0);
   set_name(0x58,"OnWindowSize",0);
   set_offset(0x5c,0,0);
   set_name(0x5c,"OnPointerMove",0);
   set_offset(0x60,0,0);
   set_name(0x60,"OnKeyDown",0);
   set_offset(0x64,0,0);
   set_name(0x64,"OnKeyUp",0);
   set_offset(0x68,0,0);
   set_name(0x68,"OnChar",0);
   set_offset(0x6c,0,0);
   set_name(0x6c,"OnTimer",0);
   set_offset(0x70,0,0);
   set_name(0x70,"OnIdle",0);
   set_offset(0x74,0,0);
   set_name(0x74,"OnMusicDone",0);
   set_offset(0x78,0,0);
   set_name(0x78,"OnQueryQuit",0);
   set_offset(0x7c,0,0);
   set_name(0x7c,"OnShutdown",0);
}

static void mark_vectors( void ) {
   set_name(get_long(0x3c),"K_ScreenChange",0);
   auto_make_code( get_long(0x3c) );
   set_name(get_long(0x40),"K_AlphaKeyDown",0);
   auto_make_code( get_long(0x40) );
   set_name(get_long(0x44),"K_AlphaKeyUp",0);
   auto_make_code( get_long(0x44) );
   set_name(get_long(0x48),"K_BeginStartup",0);
   auto_make_code( get_long(0x48) );
   set_name(get_long(0x4c),"K_EndStartup",0);
   auto_make_code( get_long(0x4c) );
   set_name(get_long(0x50),"K_WindowHidden",0);
   auto_make_code( get_long(0x50) );
   set_name(get_long(0x54),"K_WindowMove",0);
   auto_make_code( get_long(0x54) );
   set_name(get_long(0x58),"K_WindowSize",0);
   auto_make_code( get_long(0x58) );
   set_name(get_long(0x5c),"K_PointerMove",0);
   auto_make_code( get_long(0x5c) );
   set_name(get_long(0x60),"K_KeyDown",0);
   auto_make_code( get_long(0x60) );
   set_name(get_long(0x64),"K_KeyUp",0);
   auto_make_code( get_long(0x64) );
   set_name(get_long(0x68),"K_Char",0);
   auto_make_code( get_long(0x68) );
   set_name(get_long(0x6c),"K_Timer",0);
   auto_make_code( get_long(0x6c) );
   set_name(get_long(0x70),"K_Idle",0);
   auto_make_code( get_long(0x70) );
   set_name(get_long(0x74),"K_MusicDone",0);
   auto_make_code( get_long(0x74) );
   set_name(get_long(0x78),"K_QueryQuit",0);
   auto_make_code( get_long(0x78) );
   set_name(get_long(0x7c),"K_Shutdown",0);
   auto_make_code( get_long(0x7c) );
}

/*
 * Kernel event handler
 *
 * @param msgid Message ID to handle
 * @param ... Variable list of arguments
 * @return 1 on success
 */
static int notify(processor_t::idp_notify msgid, ...) {
  va_list va;
  va_start(va, msgid);

  // do IDA callbacks
  int code = invoke_callbacks(HT_IDP, msgid, va);
  if (code) return code;

  // deal with notification codes
  switch(msgid) {
  case processor_t::init:
    inf.mf = 0;       // Set little endian mode in the IDA kernel
    break;

  case processor_t::newfile: 
    {
      // create a new segment for the vector area
      add_segm(0, 0, 0x80, "vector", "CONST" );
      // create a new segment for code area
      add_segm(0, 0x80, inf.maxEA, "code", "CODE" );
      // create a new segment for data area
      inf.lowoff = inf.maxEA;
      inf.highoff = inf.maxEA + get_long(0xc); // add the zerosegment size
      add_segm(0, inf.lowoff, inf.highoff, "zerodata", "DATA" );
      mark_header( );
      mark_vectors( );
    }
    break;

  case processor_t::newseg: 
    {
 //     segment_t *newseg = va_arg(va, segment_t *);
    }
    break;

  default:
    break;
  }
  va_end(va);

  // OK
  return(1);
}

/************************************************************************/
/* Register names							*/
/************************************************************************/
static const char *RegNames[] = {
  "cs", "ds" // required by IDA kernel
};

/************************************************************************/
/* Definition of KINT assembler						*/
/************************************************************************/
static asm_t kintasm = {
  AS_COLON | ASH_HEXF3,
  0,
  "KINT Assembler",
  0,
  NULL,		// headers
  NULL,		// no bad instructions
  "; org",	// org
  "@end",	// end

  ";",		// comment string
  '\"',		// string delimiter
  '\'',		// char delimiter
  "\\\"'",	// special symbols in char+string constants

  "BYTE",	// ascii string directive
  "BYTE",	// byte directive
  "WORD",	// word directive
  "DWORD",	// dword  (4 bytes)
  NULL,		// qword  (8 bytes)
  NULL,		// oword  (16 bytes)
  NULL,		// float  (4 bytes)
  NULL,		// double (8 bytes)
  NULL,		// tbyte  (10/12 bytes)
  NULL,		// packed decimal real
  "[#d] #v",	// arrays (#h,#d,#v,#s(...)
  "BYTE [%s]",	// uninited arrays
  "EQU",	// equ
  NULL,		// seg prefix
  NULL, NULL, NULL,
  NULL,
  "$",
  kint_func_header, // func_header
  NULL,		// func_footer
  NULL,		// public
  NULL,		// weak
  NULL,		// extrn
  NULL,		// comm
  NULL,		// get_type_name
  NULL,		// align
  '(', ')',	// lbrace, rbrace
  NULL,    // mod
  NULL,    // and
  NULL,    // or
  NULL,    // xor
  NULL,    // not
  NULL,    // shl
  NULL,    // shr
  NULL,    // sizeof
  0,
};

/************************************************************************/
/* Assemblers supported by this module					*/
/************************************************************************/
static asm_t *asms[] = { &kintasm, NULL };


/************************************************************************/
/* Short names of processor						*/
/************************************************************************/
static const char *shnames[] =
{
  "kint",
  NULL
};

/************************************************************************/
/* Long names of processor						*/
/************************************************************************/
static const char *lnames[] = {
  "KINT virtual processor",
  NULL
};

//-----------------------------------------------------------------------
//	Processor Definition
//-----------------------------------------------------------------------
processor_t LPH =
{
  IDP_INTERFACE_VERSION,// version
  PLFM_KINT,	       // id
  PR_RNAMESOK|PR_USE32|PR_DEFSEG32|PR_BINMEM,   // can use register names for byte names
  8,			// 8 bits in a byte for code segments
  8,			// 8 bits in a byte for other segments

  shnames,		// array of short processor names
			// the short names are used to specify the processor
			// with the -p command line switch)
  lnames,		// array of long processor names
			// the long names are used to build the processor type
			// selection menu

  asms,			// array of target assemblers

  notify,		// the kernel event notification callback

  header,		// generate the disassembly header
  footer,		// generate the disassembly footer

  segstart,		// generate a segment declaration (start of segment)
  segfooter,		// generate a segment footer (end of segment)

  NULL,			// generate 'assume' directives

  ana,			// analyze an instruction and fill the 'cmd' structure
  emu,			// emulate an instruction

  out,			// generate a text representation of an instruction
  outop,		// generate a text representation of an operand
  out_data,		// generate a text representation of a data item
  NULL,			// compare operands
  NULL,			// can an operand have a type?

  qnumber(RegNames),	// Number of registers
  RegNames,		// Register names
  NULL,			// get abstract register

  0,			// Number of register files
  NULL,			// Register file names
  NULL,			// Register descriptions
  NULL,			// Pointer to CPU registers

  rVcs,rVds,
  0,			// size of a segment register
  rVcs,rVds,

  NULL,			// No known code start sequences
  NULL,			// return codes

  0,KINT_last,
  Instructions,

  NULL,                 // int  (*is_far_jump)(int icode);
  NULL,                 // Translation function for offsets
  0,                    // int tbyte_size;  -- doesn't exist
  NULL,                 // int (*realcvt)(void *m, ushort *e, ushort swt);
  { 0, 0, 0, 0 },       // char real_width[4];
                        // number of symbols after decimal point
                        // 2byte float (0-does not exist)
                        // normal float
                        // normal double
                        // long double
  NULL,                 // int (*is_switch)(switch_info_t *si);
  NULL,                 // int32 (*gen_map_file)(FILE *fp);
  NULL,                 // ea_t (*extract_address)(ea_t ea,const char *string,int x);
  NULL,                 // Check whether the operand is relative to stack pointer
  create_func_frame,    // create frame of newly created function
  NULL,                 // Get size of function return address in bytes
  gen_stkvar_def, 	// Generate stack variable definition line
};
